
#include "stdafx.h"

// [^̌łLɂ
#define ENABLE_TH145

// SYOLɂ
#define ENABLE_TH135

namespace {

constexpr unsigned int SIGNATURE_MASK = 0xFFFF0000;
constexpr unsigned int ARG_MASK = 0x0000FFFF;

constexpr unsigned int SIGNATURE_FIND_FILE_W = 0x00010000;
constexpr unsigned int SIGNATURE_NATIVE_FILE = 0x00020000;
constexpr unsigned int SIGNATURE_VIRTUAL_FILE = 0x00030000;

struct CloseHandleProc {
  void operator()(const HANDLE handle) const {
    ::CloseHandle(handle);
  }
};

class VirtualFileBase {
public:
  enum SeekType {
    BEGIN, CURRENT, END
  };
  virtual ~VirtualFileBase() {};
  virtual unsigned int tell() const = 0;
  virtual bool seek(unsigned int dis, const SeekType seekType) = 0;
  virtual unsigned int read(unsigned char * const buf, const unsigned int size) = 0;
};

typedef std::unique_ptr<std::remove_pointer<HANDLE>::type, CloseHandleProc> NativeFileHandle;
typedef std::unique_ptr<VirtualFileBase> VirtualFileHandle;

std::vector<std::pair<unsigned int, std::vector<WIN32_FIND_DATAW> > > findFileAryW;
std::vector<NativeFileHandle> nativeFileAry;
std::vector<VirtualFileHandle> virtualFileAry;

unsigned int getArg(const unsigned int handle) {
  return handle & ARG_MASK;
}

unsigned int getSignature(const unsigned int handle) {
  return handle & SIGNATURE_MASK;
}

unsigned int isNativeFileHandle(const unsigned int handle) {
  return getSignature(handle) == SIGNATURE_NATIVE_FILE;
}

unsigned int isVirtualFileHandle(const unsigned int handle) {
  return getSignature(handle) == SIGNATURE_VIRTUAL_FILE;
}

bool isKnownHandle(const unsigned int handle) {
  const unsigned int arg = getArg(handle);
  switch (getSignature(handle)) {
  case SIGNATURE_FIND_FILE_W:
    return arg < findFileAryW.size();
  case SIGNATURE_NATIVE_FILE:
    return arg < nativeFileAry.size();
  case SIGNATURE_VIRTUAL_FILE:
    return arg < virtualFileAry.size();
  default:
    return false;
  }
}

unsigned int newHandle(const unsigned int signature) {
  switch (signature) {
  case SIGNATURE_FIND_FILE_W:
    findFileAryW.resize(findFileAryW.size() + 1);
    return SIGNATURE_FIND_FILE_W + (findFileAryW.size() - 1);
  case SIGNATURE_NATIVE_FILE:
    nativeFileAry.resize(nativeFileAry.size() + 1);
    return SIGNATURE_NATIVE_FILE + (nativeFileAry.size() - 1);
  case SIGNATURE_VIRTUAL_FILE:
    virtualFileAry.resize(virtualFileAry.size() + 1);
    return SIGNATURE_VIRTUAL_FILE + (virtualFileAry.size() - 1);
  default:
    assert(false);
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
}

unsigned int newFindFileWHandle() {
  return newHandle(SIGNATURE_FIND_FILE_W);
}

unsigned int newNativeFileHandle() {
  return newHandle(SIGNATURE_NATIVE_FILE);
}

unsigned int newVirtualFileHandle() {
  return newHandle(SIGNATURE_VIRTUAL_FILE);
}

template<typename CONTAINER, typename IS_EMPTY_PROC, typename CLEAR_PROC>
void closeAnyHandle(const unsigned int arg, CONTAINER &container, const IS_EMPTY_PROC &isEmptyProc, const CLEAR_PROC &clearProc) {
  const unsigned int curSize = container.size();
  if (arg < curSize - 1) {
    clearProc(container[arg]);
    return;
  }
  const std::reverse_iterator<CONTAINER::iterator> begin(container.end());
  const std::reverse_iterator<CONTAINER::iterator> end(container.begin());
  const std::reverse_iterator<CONTAINER::iterator> it = std::find_if(
    begin + 1,
    end,
    [isEmptyProc](const auto &item) {
      return !isEmptyProc(item);
    }
  );
  container.resize(container.size() - std::distance(begin, it));
}

void closeFindFileW(const unsigned int arg) {
  closeAnyHandle(
    arg,
    findFileAryW,
    [](const std::pair<unsigned int, std::vector<WIN32_FIND_DATAW> > &item) { return item.second.empty(); },
    [](std::pair<unsigned int, std::vector<WIN32_FIND_DATAW> > &item) { item.second.clear(); });
}

void closeNativeFileHandle(const unsigned int arg) {
  closeAnyHandle(arg, nativeFileAry, [](const NativeFileHandle &item) { return !item; }, [](NativeFileHandle &item) { item.reset(); });
}

void closeVirtualFileHandle(const unsigned int arg) {
  closeAnyHandle(arg, virtualFileAry, [](const VirtualFileHandle &item) { return !item; }, [](VirtualFileHandle &item) { item.reset(); });
}

void closeHandle(const unsigned int handle) {
  const unsigned int arg = getArg(handle);
  switch (getSignature(handle)) {
  case SIGNATURE_FIND_FILE_W:
    closeFindFileW(arg);
    return;
  case SIGNATURE_NATIVE_FILE:
    closeNativeFileHandle(arg);
    return;
  case SIGNATURE_VIRTUAL_FILE:
    closeVirtualFileHandle(arg);
    return;
  default:
    assert(false);
    return;
  }
}

std::pair<unsigned int, std::vector<WIN32_FIND_DATAW> > &getFindFileAryW(const unsigned int handle) {
  assert(getSignature(handle) == SIGNATURE_FIND_FILE_W);
  return findFileAryW[getArg(handle)];
}

void hookFindFirstFileW(const wchar_t * const lpFileName, std::vector<WIN32_FIND_DATAW> &data);

HANDLE __stdcall d_FindFirstFileW(const wchar_t * const lpFileName, WIN32_FIND_DATAW * const lpFindFileData) {
  const unsigned int handle = newFindFileWHandle();
  std::pair<unsigned int, std::vector<WIN32_FIND_DATAW> > &data = getFindFileAryW(handle);
  hookFindFirstFileW(lpFileName, data.second);
  WIN32_FIND_DATAW item = { 0 };
  HANDLE win32handle = INVALID_HANDLE_VALUE;
  for (win32handle = ::FindFirstFileW(lpFileName, &item); win32handle != INVALID_HANDLE_VALUE; ) {
    data.second.push_back(item);
    if (::FindNextFileW(win32handle, &item) == FALSE) {
      break;
    }
  }
  ::FindClose(win32handle);
  if (data.second.empty()) {
    closeHandle(handle);
    return INVALID_HANDLE_VALUE;
  }
  data.first = 1;
  *lpFindFileData = data.second[0];
  return reinterpret_cast<HANDLE>(handle);
}

BOOL __stdcall d_FindNextFileW(const HANDLE hFindFile, WIN32_FIND_DATAW * const lpFindFileData) {
  const unsigned int handle = reinterpret_cast<unsigned int>(hFindFile);
  std::pair<unsigned int, std::vector<WIN32_FIND_DATAW> > &data = getFindFileAryW(handle);
  if (data.first >= data.second.size()) {
    ::SetLastError(ERROR_NO_MORE_FILES);
    return FALSE;
  }
  *lpFindFileData = data.second[data.first];
  data.first++;
  return TRUE;
}

BOOL __stdcall d_FindClose(HANDLE hFindFile) {
  closeHandle(reinterpret_cast<unsigned int>(hFindFile));
  return TRUE;
}

class VirtualFileOnMemory : public VirtualFileBase {
private:
  const std::vector<unsigned char> data;
  unsigned int pos;

public:
  VirtualFileOnMemory(std::vector<unsigned char> data) : data(std::move(data)), pos(0) {
  }
  ~VirtualFileOnMemory() override {
  }

  unsigned int tell() const override {
    return pos;
  }

  bool seek(unsigned int dis, const SeekType seekType) override {
    unsigned int ptr = 0;
    switch (seekType) {
    case BEGIN:
      ptr = dis;
      break;
    case CURRENT:
      ptr = pos + dis;
      break;
    case END:
      ptr = data.size() + dis;
      break;
    default:
      return false;
    }
    if (ptr < 0) {
      ::SetLastError(ERROR_NEGATIVE_SEEK);
      return false;
    }
    if (ptr > data.size()) {
      ptr = data.size();
    }
    pos = ptr;
    return true;
  }

  unsigned int read(unsigned char * const buf, const unsigned int size) override {
    unsigned int readSize = size;
    if (data.size() < pos + size) {
      readSize = data.size() - pos;
      ::SetLastError(ERROR_HANDLE_EOF);
    }
    std::copy(
      data.begin() + pos,
      data.begin() + pos + readSize,
      stdext::checked_array_iterator<unsigned char *>(buf, readSize));
    pos += readSize;
    return readSize;
  }

};

NativeFileHandle &getNativeFileHandle(const unsigned int handle) {
  assert(getSignature(handle) == SIGNATURE_NATIVE_FILE);
  return nativeFileAry[getArg(handle)];
}

VirtualFileHandle &getVirtualFileHandle(const unsigned int handle) {
  assert(getSignature(handle) == SIGNATURE_VIRTUAL_FILE);
  return virtualFileAry[getArg(handle)];
}

unsigned int hookCreateFileW(const wchar_t * const lpFileName, const DWORD dwDesiredAccess, const DWORD dwCreationDisposition);

HANDLE __stdcall d_CreateFileW(
  const wchar_t * const lpFileName,
  const DWORD dwDesiredAccess,
  const DWORD dwShareMode,
  SECURITY_ATTRIBUTES * const lpSecurityAttributes,
  const DWORD dwCreationDisposition,
  const DWORD dwFlagsAndAttributes,
  const HANDLE hTemplateFile)
{
  const unsigned int virtualHandle = hookCreateFileW(lpFileName, dwDesiredAccess, dwCreationDisposition);
  if (reinterpret_cast<HANDLE>(virtualHandle) != INVALID_HANDLE_VALUE) {
    return reinterpret_cast<HANDLE>(virtualHandle);
  }
  const HANDLE rawHandle = ::CreateFileW(lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile);
  if (rawHandle == INVALID_HANDLE_VALUE) {
    return INVALID_HANDLE_VALUE;
  }
  const unsigned int result = newNativeFileHandle();
  NativeFileHandle &file = getNativeFileHandle(result);
  file.reset(rawHandle);
  return reinterpret_cast<HANDLE>(result);
}

BOOL __stdcall d_CloseHandle(const HANDLE rawHandle) {
  const unsigned int handle = reinterpret_cast<unsigned int>(rawHandle);
  if (isKnownHandle(handle)) {
    closeHandle(handle);
    return TRUE;
  }
  return ::CloseHandle(rawHandle);
}

BOOL __stdcall d_ReadFile(const HANDLE hFile, void * const lpBuffer, const DWORD nNumberOfBytesToRead, DWORD * const lpNumberOfBytesRead, OVERLAPPED * const lpOverlapped) {
  const unsigned int handle = reinterpret_cast<unsigned int>(hFile);
  if (isNativeFileHandle(handle)) {
    return ::ReadFile(getNativeFileHandle(handle).get(), lpBuffer, nNumberOfBytesToRead, lpNumberOfBytesRead, lpOverlapped);
  }
  assert(isVirtualFileHandle(handle));
  VirtualFileHandle &file = getVirtualFileHandle(handle);
  if (lpBuffer == nullptr) {
    return FALSE;
  }
  const unsigned int readSize = file->read(reinterpret_cast<unsigned char *>(lpBuffer), nNumberOfBytesToRead);
  if (lpNumberOfBytesRead != nullptr) {
    *lpNumberOfBytesRead = readSize;
  }
  return TRUE;
}

DWORD __stdcall d_SetFilePointer(const HANDLE hFile, const LONG lDistanceToMove, LONG * const lpDistanceToMoveHigh, const DWORD dwMoveMethod) {
  const unsigned int handle = reinterpret_cast<unsigned int>(hFile);
  if (isNativeFileHandle(handle)) {
    return ::SetFilePointer(getNativeFileHandle(handle).get(), lDistanceToMove, lpDistanceToMoveHigh, dwMoveMethod);
  }
  assert(isVirtualFileHandle(handle));
  VirtualFileHandle &file = getVirtualFileHandle(handle);
  long long int distance = 0;
  if (lpDistanceToMoveHigh == nullptr) {
    distance = lDistanceToMove;
  } else {
    distance = static_cast<unsigned int>(lDistanceToMove) | (static_cast<long long int>(*lpDistanceToMoveHigh) << 32);
  }
  unsigned long long int ptr = 0;
  VirtualFileBase::SeekType type;
  switch (dwMoveMethod) {
  case FILE_BEGIN:
    type = VirtualFileBase::BEGIN;
    break;
  case FILE_CURRENT:
    type = VirtualFileBase::CURRENT;
    break;
  case FILE_END:
    type = VirtualFileBase::END;
    break;
  default:
    return FALSE;
  }
  return file->seek(static_cast<unsigned int>(distance), type) ? TRUE : FALSE;
}

DWORD __stdcall d_GetFileType(HANDLE hFile) {
  const unsigned int handle = reinterpret_cast<unsigned int>(hFile);
  if (isNativeFileHandle(handle)) {
    return ::GetFileType(getNativeFileHandle(handle).get());
  }
  if (isVirtualFileHandle(handle)) {
    return FILE_TYPE_DISK;
  }
  return ::GetFileType(hFile);
}

BOOL __stdcall d_WriteFile(
  const HANDLE hFile,
  const void * const lpBuffer,
  const DWORD nNumberOfBytesToWrite,
  DWORD * const lpNumberOfBytesWritten,
  OVERLAPPED * const lpOverlapped)
{
  const unsigned int handle = reinterpret_cast<unsigned int>(hFile);
  if (isNativeFileHandle(handle)) {
    return ::WriteFile(getNativeFileHandle(handle).get(), lpBuffer, nNumberOfBytesToWrite, lpNumberOfBytesWritten, lpOverlapped);
  }
  assert(isVirtualFileHandle(handle));
  assert(false);
  return FALSE;
}

bool hookPathFileExistsW(const wchar_t * const pszPath);

BOOL __stdcall d_PathFileExistsW(const wchar_t * const pszPath) {
  return ::PathFileExistsW(pszPath) || hookPathFileExistsW(pszPath);
}


// _~[܂
// Ǝ@\


constexpr wchar_t LIST_DIR_NAME[] = L"list";

constexpr wchar_t TH135_LIST_FILENAME[] = L"135.txt";
constexpr wchar_t TH145_LIST_FILENAME[] = L"145.txt";

struct MusicInfo {
  const char * const path;
  const unsigned int headerSize;
  const unsigned int introSize;
  const unsigned int bodySize;
  const char * const name;
};

const MusicInfo th135MusicList[] = {
  { "data/bgm/reimu1.ogg", 0x00000000, 0x0004927c * 4, 0x00461595 * 4, "tFa@` Colorful Path" },
  { "data/bgm/marisa1.ogg", 0x00000000, 0x0003ca1d * 4, 0x0050bfe0 * 4, "CKXiCg" },
  { "data/bgm/ichirin1.ogg", 0x00000000, 0x000159cc * 4, 0x0046e14c * 4, "eƃnCJ" },
  { "data/bgm/hijiri1.ogg", 0x00000000, 0x00098cee * 4, 0x00546b7c * 4, "̖VO@` Cosmic Mind" },
  { "data/bgm/futo1.ogg", 0x00000000, 0x00031a05 * 4, 0x005237bf * 4, "__b`" },
  { "data/bgm/miko1.ogg", 0x00000000, 0x001999a3 * 4, 0x004fcf7e * 4, "`@` True Administrator" },
  { "data/bgm/nitori1.ogg", 0x00000000, 0x00074443 * 4, 0x004ae670 * 4, "H열V͓̉@` Candid Friend" },
  { "data/bgm/koishi1.ogg", 0x00000000, 0x0003e4d4 * 4, 0x004f5719 * 4, "ng}̗d" },
  { "data/bgm/mamizou1.ogg", 0x00000000, 0x000367dc * 4, 0x0050c0a8 * 4, "ñjb" },
  { "data/bgm/chuboss1.ogg", 0x00000000, 0x00097c77 * 4, 0x004c6c8c * 4, "z̓b" },
  { "data/bgm/kokoro1.ogg", 0x00000000, 0x0007b749 * 4, 0x00a8b328 * 4, "S̃G[V" },
  { "data/bgm/op.ogg", 0x00000000, 0x00000000 * 4, 0x001fd6b2 * 4, "oEsς̃yV~Y" },
  { "data/bgm/title.ogg", 0x00000000, 0x00030600 * 4, 0x00409982 * 4, "SYOq" },
  { "data/bgm/ed.ogg", 0x00000000, 0x00025519 * 4, 0x00307320 * 4, "ŉ_" },
  { "data/bgm/staffroll.ogg", 0x00000000, 0x00000000 * 4, 0x003053f2 * 4, "V" },
  { "data/bgm/win.ogg", 0x00000000, 0x0006bcae * 4, 0x000e5af3 * 4, "{̈ʋL" },
  //{ "data/bgm/staffroll2.ogg", 0x00000000, 0x00000000 * 4, 0x0036162b * 4, "" },
  { "data/bgm/kaiwa1.ogg", 0x00000000, 0x0001d86e * 4, 0x002b1101 * 4, "lĈꏊ" },
  { "data/bgm/kaiwa3.ogg", 0x00000000, 0x0001e847 * 4, 0x002b10fe * 4, "lĈȂꏊ" },
  //{ "data/bgm/kaiwa4.ogg", 0x00000000, 0x00037aa4 * 4, 0x000e5b13 * 4, "" },
  { "data/bgm/kaiwa5.ogg", 0x00000000, 0x0006219a * 4, 0x002241e1 * 4, "NO̗" },
  { "data/bgm/charasel.ogg", 0x00000000, 0x0005027f * 4, 0x003126e4 * 4, "ґI" },
  { "data/bgm/reimu2.ogg", 0x00000000, 0x00000000 * 4, 0x000e7742 * 4, "lC^얲" },
  { "data/bgm/marisa2.ogg", 0x00000000, 0x00000000 * 4, 0x000e7780 * 4, "lC^J" },
  { "data/bgm/ichirin2.ogg", 0x00000000, 0x00000000 * 4, 0x000e779b * 4, "lC^_ց_R" },
  { "data/bgm/hijiri2.ogg", 0x00000000, 0x00000000 * 4, 0x000e778d * 4, "lC^@" },
  { "data/bgm/futo2.ogg", 0x00000000, 0x00000000 * 4, 0x000e778c * 4, "lC^zs" },
  { "data/bgm/miko2.ogg", 0x00000000, 0x00000000 * 4, 0x000e778a * 4, "lC^L_q" },
  { "data/bgm/nitori2.ogg", 0x00000000, 0x00000000 * 4, 0x000e7797 * 4, "lC^͏ɂƂ" },
  { "data/bgm/koishi2.ogg", 0x00000000, 0x00000000 * 4, 0x000e7769 * 4, "lC^Ön" },
  { "data/bgm/mamizou2.ogg", 0x00000000, 0x00000000 * 4, 0x000e777b * 4, "lC^b}~]E" },
  { "data/bgm/kokoro2.ogg", 0x00000000, 0x00000000 * 4, 0x000e777f * 4, "lC^`" },
  { "data/bgm/maxwaza.ogg", 0x00000000, 0x00029679 * 4, 0x001c2b1d * 4, "Xg[h" },
};

const MusicInfo th145MusicList[] = {
  { "data/bgm/title.ogg", 0x00000000, 0x000395f1 * 4, 0x003c8fe8 * 4, "^Cg" },
  { "data/bgm/st01.ogg", 0x00000000, 0x0004c9dd * 4, 0x004b30d5 * 4, "̌Stage2(_)" },
  { "data/bgm/st02.ogg", 0x00000000, 0x00059146 * 4, 0x00435e3e * 4, "̌Stage1()" },
  { "data/bgm/ta02.ogg", 0x00000000, 0x0008d314 * 4, 0x00233c76 * 4, "bStage1" },
  { "data/bgm/ta01.ogg", 0x00000000, 0x000564d3 * 4, 0x002b1100 * 4, "bStage2" },
  { "data/bgm/ta03.ogg", 0x00000000, 0x00000000 * 4, 0x001274ff * 4, "bStage31" },
  { "data/bgm/ta04.ogg", 0x00000000, 0x0001efe6 * 4, 0x00081330 * 4, "bStage32" },
};

std::string createMusicInfoString(const MusicInfo * const info, const unsigned int count) {
  return count == 0 ? "" : (
    (boost::format("\n%%%s,%08x,%08x,%08x,%s") % info[0].path % info[0].headerSize % info[0].introSize % info[0].bodySize % info[0].name).str()
    + createMusicInfoString(&info[1], count - 1)
  );
}

std::string createTh135List() {
  return std::string("@,SYO") + createMusicInfoString(th135MusicList, _countof(th135MusicList));
}

std::string createTh145List() {
  return std::string("@,[^̌") + createMusicInfoString(th145MusicList, _countof(th145MusicList));
}

const std::string th135ListTxt = createTh135List();
const std::string th145ListTxt = createTh145List();

boost::filesystem::path getAppDir() {
  wchar_t buf[4096];
  ::GetModuleFileNameW(NULL, buf, _countof(buf));
  const boost::filesystem::path exePath = buf;
  boost::filesystem::path exeDir = buf;
  exeDir.remove_filename();
  return exeDir;
}

boost::filesystem::path getListDir() {
  return getAppDir() / LIST_DIR_NAME;
}

boost::filesystem::path getTh135ListPath() {
  return getListDir() / TH135_LIST_FILENAME;
}

boost::filesystem::path getTh145ListPath() {
  return getListDir() / TH145_LIST_FILENAME;
}

void hookFindFirstFileW(const wchar_t * const lpFileName, std::vector<WIN32_FIND_DATAW> &data) {
  const boost::filesystem::path targetPath = getListDir() / L"*.txt";
  if (targetPath != lpFileName) {
    return;
  }
  struct {
    const wchar_t * const fileName;
    const unsigned int fileSize;
  } list[] = {
#ifdef ENABLE_TH135
    { TH135_LIST_FILENAME, th135ListTxt.size() },
#endif
#ifdef ENABLE_TH145
    { TH145_LIST_FILENAME, th145ListTxt.size() },
#endif
  };
  data.resize(data.size() + _countof(list));
  auto it = data.end() - _countof(list);
  SYSTEMTIME st;
  ::GetSystemTime(&st);
  for (const auto &item : list) {
    WIN32_FIND_DATAW &fd = *it;
    fd.dwFileAttributes = FILE_ATTRIBUTE_READONLY;
    ::SystemTimeToFileTime(&st, &fd.ftCreationTime);
    ::SystemTimeToFileTime(&st, &fd.ftLastAccessTime);
    ::SystemTimeToFileTime(&st, &fd.ftLastWriteTime);
    fd.nFileSizeLow = item.fileSize;
    ::wcscpy_s(fd.cFileName, _countof(fd.cFileName), item.fileName);
    it++;
  }
}

unsigned int hookCreateFileWTh135List(const wchar_t * const lpFileName, const DWORD dwDesiredAccess, const DWORD dwCreationDisposition) {
  const boost::filesystem::path listPath = getTh135ListPath();
  if (listPath != lpFileName) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  const unsigned int handle = newVirtualFileHandle();
  VirtualFileHandle &file = getVirtualFileHandle(handle);
  file = std::make_unique<VirtualFileOnMemory>(
    std::vector<unsigned char>(th135ListTxt.begin(), th135ListTxt.end())
  );
  return handle;
}

unsigned int hookCreateFileWTh145List(const wchar_t * const lpFileName, const DWORD dwDesiredAccess, const DWORD dwCreationDisposition) {
  const boost::filesystem::path listPath = getTh145ListPath();
  if (listPath != lpFileName) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  const unsigned int handle = newVirtualFileHandle();
  VirtualFileHandle &file = getVirtualFileHandle(handle);
  file = std::make_unique<VirtualFileOnMemory>(
    std::vector<unsigned char>(th145ListTxt.begin(), th145ListTxt.end())
    );
  return handle;
}

class VirtualWavFile : public VirtualFileBase {
private:
  OggVorbis_File ovf;
  const unsigned int blockSize;
  std::vector<unsigned char> temp;

public:
  VirtualWavFile(OggVorbis_File ovf) : ovf(ovf), blockSize(4) {
  }
  ~VirtualWavFile() override {
    ::ov_clear(&ovf);
  }

  unsigned int size() const {
    return static_cast<unsigned int>(::ov_pcm_total(const_cast<OggVorbis_File *>(&ovf), -1)) * blockSize;
  }

  unsigned int tell() const override {
    return static_cast<unsigned int>(::ov_pcm_tell(const_cast<OggVorbis_File *>(&ovf))) * blockSize - temp.size();
  }

  bool seek(unsigned int dis, const SeekType seekType) override {
    assert((dis % blockSize) == 0);
    const unsigned int pos = tell();
    const unsigned int totalSize = size();
    unsigned int ptr = 0;
    switch (seekType) {
    case BEGIN:
      ptr = dis;
      break;
    case CURRENT:
      ptr = pos + dis;
      break;
    case END:
      ptr = totalSize + dis;
      break;
    default:
      return false;
    }
    if (ptr > totalSize) {
      ptr = totalSize;
    }
    ::ov_pcm_seek(&ovf, ptr / blockSize);
    temp.clear();
    return true;
  }

  unsigned int read(unsigned char * const buf, const unsigned int bufSize) override {
    const unsigned int tempSize = temp.size();
    assert((tempSize % 4) == 0);
    if (!temp.empty() && bufSize < tempSize) {
      std::copy(temp.begin(), temp.begin() + bufSize, stdext::checked_array_iterator<unsigned char *>(buf, bufSize));
      std::copy(temp.begin() + bufSize, temp.end(), temp.begin());
      temp.resize(tempSize - bufSize);
      return bufSize;
    }
    const unsigned int pos = static_cast<unsigned int>(::ov_pcm_tell(const_cast<OggVorbis_File *>(&ovf))) * blockSize;
    const unsigned int totalSize = size();
    unsigned int readSize = bufSize - tempSize;
    if (pos + readSize > totalSize) {
      readSize = totalSize - pos - tempSize;
    }
    temp.resize(tempSize + (std::max)(bufSize * 2, 4096u));
    int bitStream = 0;
    unsigned int readSizeResult = 0;
    unsigned int tempIndex = tempSize;
    while (true) {
      const unsigned int curReadSize = ::ov_read(&ovf, reinterpret_cast<char *>(&temp[tempIndex]), temp.size() - tempIndex, 0, 2, 1, &bitStream);
      if (curReadSize == 0) {
        break;
      }
      readSizeResult += curReadSize;
      tempIndex += curReadSize;
      if (tempIndex == temp.size()) {
        break;
      }
    }
    temp.resize(tempSize + readSizeResult);
    if (readSizeResult < readSize) {
      readSize = readSizeResult;
    }
    std::copy(temp.begin(), temp.begin() + tempSize + readSize, stdext::checked_array_iterator<unsigned char *>(buf, bufSize));
    std::copy(temp.begin() + tempSize + readSize, temp.end(), temp.begin());
    temp.resize(temp.size() - tempSize - readSize);
    if (readSize == 0 && tempSize == 0) {
      ::SetLastError(ERROR_HANDLE_EOF);
    }
    return tempSize + readSize;
  }
};

unsigned int hookCreateFileWNativeOgg(const wchar_t * const lpFileName, const DWORD dwDesiredAccess, const DWORD dwCreationDisposition) {
  const boost::filesystem::path path = lpFileName;
  if (path.extension() != ".ogg" || !boost::filesystem::is_regular_file(path)) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  OggVorbis_File ovf;
  if (::ov_fopen(path.string().c_str(), &ovf) != 0) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  const unsigned int handle = newVirtualFileHandle();
  VirtualFileHandle &file = getVirtualFileHandle(handle);
  file = std::make_unique<VirtualWavFile>(ovf);
  return handle;
}

template<typename T>
class ReaderBase : public boost::noncopyable {
private:
  T &pak;
  const TouhouSE::TH135::FileInfo &info;
  unsigned int ptr;

public:
  ReaderBase(T &pak, const TouhouSE::TH135::FileInfo &info) : pak(pak), info(info), ptr(0) {
    seek(0, SEEK_SET);
  }

  unsigned int size() const {
    return info.size;
  }

  unsigned int tell() const {
    return ptr;
  }

  bool seek(unsigned int dis, const int origin) {
    const unsigned int pos = tell();
    const unsigned int totalSize = size();
    unsigned int ptrPos = 0;
    switch (origin) {
    case SEEK_SET:
      ptrPos = dis;
      break;
    case SEEK_CUR:
      ptrPos = pos + dis;
      break;
    case SEEK_END:
      ptrPos = totalSize + dis;
      break;
    default:
      return false;
    }
    if (ptrPos > totalSize) {
      ptrPos = totalSize;
    }
    this->ptr = ptrPos;
    pak.in->seekg(pak.headerSize + info.address + this->ptr);
    return true;
  }

  unsigned int read(unsigned char * const buf, const unsigned int bufSize) {
    const unsigned int pos = tell();
    const unsigned int totalSize = size();
    unsigned int readSize = bufSize;
    if (pos + readSize > totalSize) {
      readSize = totalSize - pos;
    }
    if (readSize == 0) {
      return 0;
    }
    pak.in->read(reinterpret_cast<char *>(buf), readSize);
    unsigned char *it = buf;
    const unsigned char * const end = &buf[readSize];
    for (unsigned int i = (ptr % TouhouSE::TH135::BLOCK_SIZE); i != 0 && buf != end && i < TouhouSE::TH135::BLOCK_SIZE; i++) {
      *it ^= info.key[i];
      it++;
    }
    const unsigned int loopCount = (end - it) / TouhouSE::TH135::BLOCK_SIZE;
    for (unsigned int i = 0; i < loopCount; i++) {
      *reinterpret_cast<unsigned long long int *>(&it[0]) ^= *reinterpret_cast<const unsigned long long int *>(&info.key[0]);
      *reinterpret_cast<unsigned long long int *>(&it[8]) ^= *reinterpret_cast<const unsigned long long int *>(&info.key[8]);
      it += TouhouSE::TH135::BLOCK_SIZE;
    }
    for (unsigned int i = 0; it != end; i++) {
      *it ^= info.key[i];
      it++;
    }
    assert(it == end);
    ptr += readSize;
    return readSize;
  }
};

class Th135Reader : public ReaderBase<TouhouSE::TH135::Th135Extracter> {
public:
  Th135Reader(TouhouSE::TH135::Th135Extracter &pak, const TouhouSE::TH135::FileInfo &info) : ReaderBase<TouhouSE::TH135::Th135Extracter>(pak, info) {
  }
};

class Th145Reader : public ReaderBase<TouhouSE::TH145::Th145Extracter> {
public:
  Th145Reader(TouhouSE::TH145::Th145Extracter &pak, const TouhouSE::TH135::FileInfo &info) : ReaderBase<TouhouSE::TH145::Th145Extracter>(pak, info) {
  }
};

template<typename T>
bool oggOpen(OggVorbis_File &ovf, T * const readerPtr) {
  const int result = ::ov_open_callbacks(readerPtr, &ovf, NULL, 0, {
    [](void * const buf, const size_t size, const size_t n, void * const ptr) -> size_t {
      T &reader = *reinterpret_cast<T *>(ptr);
      return reader.read(reinterpret_cast<unsigned char *>(buf), size * n) / size;
    },
      [](void * const ptr, const ogg_int64_t dis, const int origin) -> int {
      T &reader = *reinterpret_cast<T *>(ptr);
      return reader.seek(static_cast<unsigned int>(dis), origin) ? 0 : 1;
    },
      [](void * const ptr) -> int {
      T * const reader = reinterpret_cast<T *>(ptr);
      delete ptr;
      return 0;
    },
      [](void * const ptr) -> long {
      const T &reader = *reinterpret_cast<T *>(ptr);
      return reader.tell();
    }
  });
  return result == 0;
}

unsigned int getRelativeRootPos(const wchar_t * const lpFileName) {
  boost::filesystem::path tempPath = lpFileName;
  while (true) {
    if (tempPath.empty()) {
      return{};
    }
    if (tempPath.filename() == L"data") {
      tempPath = tempPath.parent_path();
      break;
    }
    tempPath = tempPath.parent_path();
  }
  return tempPath.wstring().size() + 1;
}

boost::filesystem::path getRelativeRootPath(const wchar_t * const lpFileName) {
  return boost::filesystem::path(&lpFileName[0], &lpFileName[getRelativeRootPos(lpFileName)]);
}

boost::filesystem::path getRelativePath(const wchar_t * const lpFileName) {
  return boost::filesystem::path(&lpFileName[getRelativeRootPos(lpFileName)], &lpFileName[wcslen(lpFileName)]);
}

boost::shared_ptr<TouhouSE::TH135::Th135Extracter> th135pak = nullptr;
boost::shared_ptr<TouhouSE::TH135::Th135Extracter> th135pakB = nullptr;

unsigned int hookCreateFileWTh135Ogg(const wchar_t * const lpFileName, const DWORD dwDesiredAccess, const DWORD dwCreationDisposition) {
  if (!th135pak || !th135pakB || !boost::filesystem::is_regular_file(getRelativeRootPath(lpFileName) / L"th135.pak")) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  const boost::filesystem::path path = getRelativePath(lpFileName);
  const TouhouSE::TH135::FileInfo * const info = th135pak->find(path);
  const TouhouSE::TH135::FileInfo * const infoB = th135pakB->find(path);
  if (info == nullptr && infoB == nullptr) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  OggVorbis_File ovf;
  Th135Reader * const readerPtr = new Th135Reader(*(infoB ? th135pakB : th135pak), *(infoB ? infoB : info));
  if (!oggOpen(ovf, readerPtr)) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  const unsigned int handle = newVirtualFileHandle();
  getVirtualFileHandle(handle) = std::make_unique<VirtualWavFile>(ovf);
  return handle;
}

boost::shared_ptr<TouhouSE::TH145::Th145Extracter> th145pak = nullptr;
boost::shared_ptr<TouhouSE::TH145::Th145Extracter> th145pakB = nullptr;

unsigned int hookCreateFileWTh145Ogg(const wchar_t * const lpFileName, const DWORD dwDesiredAccess, const DWORD dwCreationDisposition) {
  if (!th145pak || !boost::filesystem::is_regular_file(getRelativeRootPath(lpFileName) / L"th145.pak")) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  const boost::filesystem::path path = getRelativePath(lpFileName);
  const TouhouSE::TH135::FileInfo * const info = th145pak->find(path);
  const TouhouSE::TH135::FileInfo * const infoB = (!th145pakB ? nullptr : th145pakB->find(path));
  if (info == nullptr && infoB == nullptr) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  OggVorbis_File ovf;
  Th145Reader * const readerPtr = new Th145Reader(*(infoB ? th145pakB : th145pak), *(infoB ? infoB : info));
  if (!oggOpen(ovf, readerPtr)) {
    return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
  }
  const unsigned int handle = newVirtualFileHandle();
  getVirtualFileHandle(handle) = std::make_unique<VirtualWavFile>(ovf);
  return handle;
}

unsigned int hookCreateFileW(const wchar_t * const lpFileName, const DWORD dwDesiredAccess, const DWORD dwCreationDisposition) {
  const unsigned int th135ListResult = hookCreateFileWTh135List(lpFileName, dwDesiredAccess, dwCreationDisposition);
  if (reinterpret_cast<HANDLE>(th135ListResult) != INVALID_HANDLE_VALUE) {
    return th135ListResult;
  }
  const unsigned int th145ListResult = hookCreateFileWTh145List(lpFileName, dwDesiredAccess, dwCreationDisposition);
  if (reinterpret_cast<HANDLE>(th145ListResult) != INVALID_HANDLE_VALUE) {
    return th145ListResult;
  }
  const unsigned int nativeOggResult = hookCreateFileWNativeOgg(lpFileName, dwDesiredAccess, dwCreationDisposition);
  if (reinterpret_cast<HANDLE>(nativeOggResult) != INVALID_HANDLE_VALUE) {
    return nativeOggResult;
  }
  const unsigned int th135OggResult = hookCreateFileWTh135Ogg(lpFileName, dwDesiredAccess, dwCreationDisposition);
  if (reinterpret_cast<HANDLE>(th135OggResult) != INVALID_HANDLE_VALUE) {
    return th135OggResult;
  }
  const unsigned int th145OggResult = hookCreateFileWTh145Ogg(lpFileName, dwDesiredAccess, dwCreationDisposition);
  if (reinterpret_cast<HANDLE>(th145OggResult) != INVALID_HANDLE_VALUE) {
    return th145OggResult;
  }
  return reinterpret_cast<unsigned int>(INVALID_HANDLE_VALUE);
}

bool hookOpenArchiveTh135() {
  const boost::filesystem::path th135PakPath = L"th135.pak";
  const boost::filesystem::path th135PakBPath = L"th135b.pak";
  if (!boost::filesystem::is_regular_file(th135PakPath) || !boost::filesystem::is_regular_file(th135PakBPath)) {
    return false;
  }
  if (th135pak && th135pakB) {
    return true;
  }
  std::unique_ptr<boost::filesystem::ifstream> ifs = std::make_unique<boost::filesystem::ifstream>(th135PakPath, std::ios::binary);
  th135pak = TouhouSE::TH135::Th135Extracter::Open(std::move(ifs), boost::filesystem::file_size(th135PakPath));
  if (!th135pak) {
    return false;
  }
  ifs = std::make_unique<boost::filesystem::ifstream>(th135PakBPath, std::ios::binary);
  th135pakB = TouhouSE::TH135::Th135Extracter::Open(std::move(ifs), boost::filesystem::file_size(th135PakBPath));
  return !!th135pakB;
}

bool hookOpenArchiveTh145() {
  const boost::filesystem::path th145PakPath = L"th145.pak";
  if (!boost::filesystem::is_regular_file(th145PakPath)) {
    return false;
  }
  if (th145pak) {
    return true;
  }
  std::unique_ptr<boost::filesystem::ifstream> ifs = std::make_unique<boost::filesystem::ifstream>(th145PakPath, std::ios::binary);
  th145pak = TouhouSE::TH145::Th145Extracter::Open(std::move(ifs), boost::filesystem::file_size(th145PakPath));
  if (!th145pak) {
    return false;
  }
  const boost::filesystem::path th145PakBPath = L"th145b.pak";
  if (!boost::filesystem::is_regular_file(th145PakBPath)) {
    return true;
  }
  ifs = std::make_unique<boost::filesystem::ifstream>(th145PakBPath, std::ios::binary);
  th145pakB = TouhouSE::TH145::Th145Extracter::Open(std::move(ifs), boost::filesystem::file_size(th145PakBPath));
  return true;
}

template<typename LIST, typename EXTRACTER>
bool isTargetFile(const wchar_t * const pszPath, const LIST &list, bool (*openProc)(), EXTRACTER &pakA, EXTRACTER &pakB) {
  const boost::filesystem::path path = pszPath;
  const auto endIt = &list[_countof(list)];
  const auto it = std::find_if(&list[0], endIt, [path](const auto &item) {
    return path == item.path;
  });
  if (it != endIt) {
    if (openProc()) {
      if (pakA->find(path) != nullptr || (pakB && pakB->find(path) != nullptr)) {
        return true;
      }
    }
  }
  return false;
}

bool isTh135File(const wchar_t * const pszPath) {
  return isTargetFile(pszPath, th135MusicList, hookOpenArchiveTh135, th135pak, th135pakB);
}

bool isTh145File(const wchar_t * const pszPath) {
  return isTargetFile(pszPath, th145MusicList, hookOpenArchiveTh145, th145pak, th145pakB);
}

bool hookPathFileExistsW(const wchar_t * const pszPath) {
  return isTh135File(pszPath) || isTh145File(pszPath);
}


// Ǝ@\܂


bool ChangeIAT(HMODULE module) {
  static const struct {
    const std::string dllName;
    const std::string procName;
    const unsigned int procAddress;
  } targetList[] = {
    { "KERNEL32.dll", "FindFirstFileW", reinterpret_cast<unsigned int>(d_FindFirstFileW) },
    { "KERNEL32.dll", "FindNextFileW", reinterpret_cast<unsigned int>(d_FindNextFileW) },
    { "KERNEL32.dll", "FindClose", reinterpret_cast<unsigned int>(d_FindClose) },
    { "KERNEL32.dll", "CreateFileW", reinterpret_cast<unsigned int>(d_CreateFileW) },
    { "KERNEL32.dll", "CloseHandle", reinterpret_cast<unsigned int>(d_CloseHandle) },
    { "KERNEL32.dll", "ReadFile", reinterpret_cast<unsigned int>(d_ReadFile) },
    { "KERNEL32.dll", "SetFilePointer", reinterpret_cast<unsigned int>(d_SetFilePointer) },
    { "KERNEL32.dll", "WriteFile", reinterpret_cast<unsigned int>(d_WriteFile) },
    { "KERNEL32.dll", "GetFileType", reinterpret_cast<unsigned int>(d_GetFileType) },
    { "SHLWAPI.dll", "PathFileExistsW", reinterpret_cast<unsigned int>(d_PathFileExistsW) },
  };
  unsigned char * const baseAddr = reinterpret_cast<unsigned char *>(module == NULL ? ::GetModuleHandleW(NULL) : module);
  const IMAGE_DOS_HEADER &mz = *reinterpret_cast<const IMAGE_DOS_HEADER *>(baseAddr);
  const IMAGE_NT_HEADERS32 &pe = *reinterpret_cast<const IMAGE_NT_HEADERS32 *>(baseAddr + mz.e_lfanew);
  const IMAGE_IMPORT_DESCRIPTOR *desc = reinterpret_cast<IMAGE_IMPORT_DESCRIPTOR *>(baseAddr + pe.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress);
  for (; desc->OriginalFirstThunk != 0; desc = ++desc) {
    const char * const dllName = reinterpret_cast<const char *>(baseAddr + desc->Name);
    const unsigned int *in = reinterpret_cast<unsigned int *>(baseAddr + desc->OriginalFirstThunk);
    unsigned int *out = reinterpret_cast<unsigned int *>(baseAddr + desc->FirstThunk);
    for (; *in != 0; ++in, ++out) {
      if ((*in & 0x80000000) != 0) {
        continue;
      }
      const char * const procName = reinterpret_cast<const char *>(baseAddr + *in + 2);
      for (const auto &target : targetList) {
        if (::_stricmp(target.dllName.c_str(), dllName) != 0 || target.procName != procName) {
          continue;
        }
        DWORD oldProtect;
        ::VirtualProtect(out, 4, PAGE_READWRITE, &oldProtect);
        *out = target.procAddress;
        ::VirtualProtect(out, 4, oldProtect, &oldProtect);
      }
    }
  }
  return true;
}

} // anonymous

void main() {
  // debugp
  //org::click3::DllHackLib::SetupConsole();
  const bool changeResult = ChangeIAT(NULL);
}
